#ifndef COLLISION_H
#define COLLISION_H

#include "Particle.h"
//#include "Object.h"
#include <iostream>
#include "Vector3.h"
#include <vector>

using std::cerr;
using std::endl;

namespace Penguin
{
	class Object;
	enum COL_TYPE {COL_SPHERE, COL_AABB, COL_CYLINDER, COL_ORFACE, COL_PLANE};

	/* Collision Structures */

	/*! \brief Abstract class for a collidable object.
	*
	* Any implementation of ColBase will normally attach itself to a particle (or a set of particles), and directly
	* influence that particle/set by providing a specification of collidable volume.
	*/	
	class ColBase
	{
		protected:
	
		/*! Can this object be moved as a result of collision? */
		bool m_Moveable;
		/*! Coefficient of restitution. */
		double m_CoefficientOfRestitution;		
		/*! Coefficient of friction. */
		double m_COF;

		public:
		
		ColBase():m_Moveable(true),m_CoefficientOfRestitution(0.55),m_COF(0.5){}
		
		/*! Returns the type of object. */
		virtual COL_TYPE GetType() = 0;
		
		double GetCoefficientOfRestitution(){return m_CoefficientOfRestitution;}
		void SetCoefficientOfRestitution(double cor){m_CoefficientOfRestitution=cor;}
		
		double GetCOF(){return m_COF;}
		void SetCOF(double cof){m_COF=cof;}
		
		bool IsMoveable(){return m_Moveable;}
		void SetMoveable(bool t){m_Moveable = t;}
	};
	
	/*! Extra interface for single particle collidable classes. */
	class ColBaseSP: public ColBase
	{
		protected:

		/*! The particle to which it is attached. */
		Particle* m_Target;

		public:

		ColBaseSP(Particle* target):ColBase(),m_Target(target){}

		/*! Get the target particle. */
		Particle* GetTarget(){return m_Target;}
		/*! Set the target particle. */	
		void SetTarget(Particle* p){m_Target = p;}
		
		virtual COL_TYPE GetType() = 0;
	};

	
	/*! \brief A Collidable Sphere. */	
	class ColSphere : public ColBaseSP
	{
		protected:
	
		/*! The sphere radius. */
		REAL m_Radius;
		
		public:
		
		ColSphere(Particle* target, double radius)
		:ColBaseSP(target), m_Radius(radius){}
				
		virtual COL_TYPE GetType(){return COL_SPHERE;}
		
		/*! Get the sphere radius. */
		REAL GetRadius(){return m_Radius;}
		/*! Sets the sphere radius. */
		void SetRadius(REAL r){m_Radius = r;}
	};

	/*! \brief A collidable plane.
	*
	* This collidable object is a static plane in the world. It is not attached to a particle and therefore
	* is stationary.
	*/
	class ColPlane : public ColBase
	{		
		protected:
	
		/*! Surface normal. */
		Vector3 m_Normal;
		/*! Some point on the plane. */
		Vector3 m_Pos;

		public:

		ColPlane(Vector3 normal, Vector3 pos)
		:ColBase(), m_Normal(normal), m_Pos(pos)
		{
			//This shape is not moveable by default
			m_Moveable = false;
			// m_Normal is a unit vector by default
			m_Normal.Normalise();
		}
	
		virtual COL_TYPE GetType(){return COL_PLANE;}
		
		/*! Get the position of the planar point. */
		Vector3& GetPos(){return m_Pos;}
		/*! Set a new position. */
		void SetPos(Vector3& v){m_Pos = v;}
		
		/*! Get the surface normal. */
		Vector3& GetNormal(){return m_Normal;}
		/*! Set the surface normal. */
		void SetNormal(Vector3& n){m_Normal = n;}
	};

	//Note: m_pos inherited from ColPlane now represents the top-left
	//corner of the face and the normal vector goes out of the top-left
	//corner as well. ColOrientedFace is also not moveable
	class ColOrientedFace : public ColPlane
	{
		protected:
	
		REAL m_width;
		REAL m_height;
		//The top edge of the face, goes out of the top-left corner
		Vector3 m_top;

		public:	

		ColOrientedFace(Vector3 pos, Vector3 normal, Vector3 t, REAL w, REAL h)
		: ColPlane(normal, pos), m_width(w), m_height(h), m_top(t)
		{
			//m_top is normalised by default
			m_top.Normalise();				
			
			//m_top must be perpendicular to the normal vector
			if (fabs(m_Normal.Dot(m_top)) > 0.01)
				{
					cerr << "Warning: an ColOrientedFace-typed object is not defined correctly."<<endl;}
		}

		virtual COL_TYPE GetType(){return COL_ORFACE;}

		REAL GetWidth(){return m_width;}
		void SetWidth(REAL w){m_width = w;}

		REAL GetHeight(){return m_height;}
		void SetHeight(REAL h){m_height = h;}

		Vector3& GetTop(){return m_top;}
		void SetTop(Vector3& t){m_top = t;}

	};
	
	class ColCylinder: public ColBase
	{
		protected:
			
		Vector3 m_top;//Centre of the top face
		Vector3 m_bottom;//Centre of the bottom face
		REAL m_radius;

		public:
	
		ColCylinder(Vector3 t, Vector3 b, REAL r)
		: ColBase(), m_top(t), m_bottom(b), m_radius(r)
		{
			//Cylinder is not moveable for now
			m_Moveable = false;
		}

		virtual COL_TYPE GetType(){return COL_CYLINDER;}

		Vector3& GetTop(){return m_top;}
		void SetTop(Vector3& t){m_top = t;}

		Vector3& GetBottom(){return m_bottom;}
		void SetBottom(Vector3& b){m_bottom = b;}

		REAL GetRadius(){return m_radius;}
		void SetRadius(REAL r){m_radius = r;}
	};

	//AABB = Axis-Aligned Bounding Box
	//AABB is not moveable
	class ColAABB: public ColBase
	{	
		protected:

		Vector3 m_centre; //The absolute centre of the AABB box
		REAL m_width;
		REAL m_height;
		REAL m_depth;
		Vector3 t1, t2, t3, t4, b1, b2, b3, b4;

		private:

		void InitPoints();

		public:

		ColAABB(Vector3 c, REAL w, REAL h, REAL d)
		: ColBase(), m_centre(c), m_width(w), m_height(h), m_depth(d)
		{
			m_Moveable = false;		
			InitPoints();
		}

		virtual COL_TYPE GetType(){return COL_AABB;}
	
		Vector3& GetCentre(){return m_centre;}
		void SetCentre(Vector3& c){m_centre = c;}

		REAL GetWidth(){return m_width;}
		void SetWidth(REAL w){m_width = w;}

		REAL GetHeight(){return m_height;}
		void SetHeight(REAL h){m_height = h;}
	
		REAL GetDepth(){return m_depth;}
		void SetDepth(REAL d){m_depth = d;}

		std::vector<Vector3*> GetFrontFace();
		std::vector<Vector3*> GetBackFace();
		std::vector<Vector3*> GetTopFace();
		std::vector<Vector3*> GetBottomFace();
		std::vector<Vector3*> GetLeftFace();
		std::vector<Vector3*> GetRightFace();
	};

	/* Collision Detection Routines */

	/*! \brief A Collision Information Structure
	*
	* This structure contains information regarding two objects (ColBase) that
	* have been checked for collision.
	*/
	struct CollisionInfo
	{
		/*! Are the objects intersecting? */		
		bool areIntersecting;				
		/*! The types of objects */
		COL_TYPE type1,type2;		
		/*! Pointers to the objects */
		ColBase *target1, *target2;		
		/*! Pointers to the Object to which the ColBase belongs */
		Object *parent; 
		
		/*! Relative velocity of objects */
		Vector3 relativeVelocity;
		/*! Relative velocity projected onto the collision normal */
		REAL velocityProjectionOnDist;		
		/*! The collision normal */
		Vector3 collisionNormal;
		/*! The distance of intersection (in the direction of collision Normal) */
		REAL intersectionDistance;
	};

	/*! Detects collision between two spheres */
	CollisionInfo DetectCollision(ColSphere* s1, ColSphere* s2, double dt);		
	/*! Detects collision between a sphere and a plane */
	CollisionInfo DetectCollision(ColPlane* p1, ColSphere* s2);
	/*! Detect collision between a sphere and an oriented face.
 	* Return a report of the nature of the collision in a Collision Info struct.
	*/
	CollisionInfo DetectCollision(ColSphere* s, ColOrientedFace* o);
	/*! Detect collision between a sphere and a cylinder.
 	* Return a report of the nature of the collision in a Collision Info struct.
	*/
	CollisionInfo DetectCollision(ColSphere* s, ColCylinder* c);
	/*! Detect collision between a sphere and an AABB.
 	* Return a report of the nature of the collision in a Collision Info struct.
	*/
	CollisionInfo DetectCollision(ColSphere* s, ColAABB* a);
	

	/*! Detect ovelapping between 2 AABB's. */
	bool Overlap(ColAABB* a1, ColAABB* a2);
	
}

#endif
