#ifndef EMITTER_H
#define EMITTER_H

#include "Object.h"

namespace Penguin
{
	/*! /brief The interface that dictates an emitters behaviour. */
	class EmitterBehaviour
	{
		public:
		
		virtual void Emit(Particle* p) = 0;
	};

	/*! /brief Emits particles over the surface of a bounded plane.
	*
	* NOTE: At the moment the plane is an implicit y-plane. With the normal downwards.
	*/
	class PlanarEmitter: public EmitterBehaviour
	{
		REAL m_WidthOn2;
		REAL m_DepthOn2;
		Vector3 m_Center;
		
		public:

		PlanarEmitter(Vector3 center, REAL width, REAL depth):m_WidthOn2(width/2),m_DepthOn2(depth/2),m_Center(center){};
		virtual void Emit(Particle* p)
		{
			REAL& x = m_Center[0];
			REAL& z = m_Center[2];
			REAL xpos = RandomREAL(x-m_WidthOn2,x+m_WidthOn2);
			REAL zpos = RandomREAL(z-m_DepthOn2,z+m_DepthOn2);
			p->SetPos(Vector3(xpos,m_Center[1],zpos));
		}
	};

	/*! /brief Emits particles like a fountain would. 
	* At the moment only sprays upwards!
	*/
	class FountainEmitter: public EmitterBehaviour
	{
		Vector3 _center;
		REAL _vel; //avg velocity along normal direction
		REAL _spray; //a radius of spray		
		
		public:

		Vector3 GetCenter(){return _center;}
		
		FountainEmitter(Vector3 center, REAL avgvel, REAL spray):_center(center),_vel(avgvel),_spray(spray){}
		
		virtual void Emit(Particle* p)
		{
				p->SetPos(_center);

				REAL r = RandomREAL(0,_spray);
				REAL theta = RandomREAL(0,2*3.14);
				
				REAL x = r * sin(theta);
				REAL z = r * cos(theta);
				
				Vector3 vel= Vector3(x,1,z)*_vel;				
				p->SetVel(vel);
		}	
	};

	/*! \brief An Emitter class.	*/	
	class Emitter: public Object
	{
		protected:
	
		EmitterBehaviour* m_EmitterBehaviour;
	
		int m_NumParticles;
		int m_CurrentParticles;

		double m_Rate; /* seconds per particle */
		double m_TimeSinceLastEmit;
		double m_InitialLife;

		double* m_LifeArray;
		void Emit(int num);
	
		double m_mass; //mass of a particle
		
		public:
		
		Emitter(int max_num, double mass);
		virtual ~Emitter();

		virtual void Update(double dt);
		virtual PIter GetParticlesEnd();
			
		/* get/sets */
		void SetRate(double rate){m_Rate = rate;}
		double GetRate(){return m_Rate;}

		void SetBehaviour(EmitterBehaviour* em)
		{
			delete m_EmitterBehaviour;
			m_EmitterBehaviour=em;
		}
		EmitterBehaviour& GetEmitterBehaviour(){return *m_EmitterBehaviour;}
		
		void SetLife(double life){m_InitialLife = life;}
		double GetLife(){return m_InitialLife;}
	
		virtual void Setup();
		virtual void UpdateParticles(double dt);
		
	};
	
	/*! \brief A collidable particle emitter.
	*
	*
	*/	
	class CollidableEmitter: public Emitter
	{
		protected:
		REAL m_radius;
	
		public:
	
		virtual ~CollidableEmitter(){};	
		CollidableEmitter(int max_num, double mass, REAL radius);
		virtual void Setup();
		virtual ColIter GetCollisionStructsEnd();
		virtual void UpdateParticles(double dt);
	};
}
#endif
