#ifndef PSYSTEM_H
#define PSYSTEM_H

#include "World.h"
#include "MPWorld.h"

#include "DampedSpring.h"
#include "ParametricParticle.h"

#include <map>
#include <string>

namespace Penguin
{
	/*! \brief A PSystem is a universe that contains Worlds.
	 *
	 * Ideally only one PSystem should exist per application. 
	 *
	 * A PSystem:
	 *  - maintains the current time passed
	 *  - updates its worlds
	 *  - manages erroneous information
	 *  - is responsible for (automatic) destruction of worlds
	 *
	 */
	 
	class PSystem
	{
		private:
		
		/*! The Worlds are stored with string keys. Allowing worlds to 
		 * be more easily identified/managed. */
		std::map<std::string,World*> m_Worlds;
		
		/*! The amount of time passed since creation. */
		double m_Time;
		
		public:
		
		PSystem();
		~PSystem();
		
		/*! Add a new world to the System. */
		World* AddWorld(std::string	name, World* world);		
		
		/*! Retrieve a world from the System. */
		World* GetWorld(std::string name);
		
		/*! Retrieve a world from the System (shorthand)*/		
		World* operator[](std::string);		
		
		/*! Remove a world from the System*/
		void RemoveWorld(std::string name);
		
		/*! Step the simulation 
		 *
		 * This functions steps each World an amount -dt-. Each World is then
		 * responsible for reaching this timestep.
		*/
		void Simulate(double dt);
		
		/*! Retrieve the time (in seconds) since epoch of the simulation. */
		double GetTime();
	};
}

#endif
