#ifndef WORLD_H
#define WORLD_H

#include "Particle.h"
#include "ParametricParticle.h"
#include "Force.h"
#include "Constraint.h"
#include "Object.h"
#include "Common.h"
#include "Collision.h"

#include <vector>
#include <map>
#include <utility>

namespace Penguin
{ 
	class PSystem;
	class CollisionHandler;
	class DefaultHandler;
	/*! \brief A World is a collection of interacting particles. 
	 *
	 * A World is a subdivision of a System. Each World contains its own
	 * interacting particles, forces and constraints.
	 * 
	 * Uses for multiple Worlds in a system include:
	 *  - Seperating non-collidible groups of particles (ie smoke and a spring).
	 *  - Giving different levels of accuracy to different systems. For example
	 *		allowing smoke particles to be simply modelled, while simultaneously
	 *    modelling a spring with high accuracy.
	 *
	 * A World is responsible for managing all the objects it contains. When a 
	 * world is removed from the System it will delete all the objects it refers
	 * to. Thus inter-world forces and constraints are not recommended.
	 *
	*/		
	class World
	{
		protected:

		PSystem* m_Parent;

		/*! A map of all Objects in this world. */
		std::map<std::string,Object*> m_Objects;

		/*! Impulse force vector */
		std::vector<ImpulseForce> m_ImpulseVector;

		/*! Force caused by gravity in the <b>-y</b> direction */
		double m_Gravity;
		
		/*! World Resistance. I.e. how much air resistance.
		*
		* Provides a resisting force to all particles
		* proportional to its speed.
		*
		*/
		double m_Resistance;
				
		/*! Collision Tolerance
		*
		* This value determines how far two objects can overlap before
		* being classed as colliding or in contact.
		*/		
		double m_CollisionTolerance;		
	
		/*! CollisionHandler vector: collisions between two Objects */
		std::map<std::pair<Object*,Object*>,CollisionHandler*>m_ObjObjCollisionHandlers;		

		/*! CollisionHandler vector: collisions between one Object and any other Object*/
		std::map<Object*,CollisionHandler*> m_ObjCollisionHandlers;

		
		/* These functions apply to all the elements in this world. */
		virtual void ZeroForces();
		virtual void ApplyForces();
		virtual void ApplyImpulses(double dt);
		virtual void SatisfyConstraints();
		virtual void ApplyEulerStep(double dt);
		virtual void SimulateParametricParticles(double t);
		virtual void UpdateObjects(double dt);
		
		/*This function applies to all collision structures*/
		void ApplyCollision(double dt);
		
		/* These functions apply to a single particle. */
		void ZeroForce(Particle* p);
		static void EulerStep(Particle* p,double dt);		
	
		/* Default collision handler */
		DefaultHandler* default_handler;
		
		public:
		
		/*! Constructor. */
		World(double gravity=0, double resistance=0.01);
		/*! Destructor. Note: See World Summary. */
		virtual ~World();
		
		/*! Set the PSystem parent. Called by the PSystem. */
		void SetParent(PSystem* p){m_Parent=p;}
		/*! This function applies to the CollisionHandler vector for collisions between two Objects */
		void RegisterObjToObjHandler(Object *o1, Object *o2, CollisionHandler *handler);
	
		/*! This function applies to the CollisionHandlr vector for colisions between an Oject and any other Object */
		void RegisterObjHandler(Object *o, CollisionHandler * handler);

		
		/*! Add a new Object with a name to this world.
		*
		* Adds an object to the world. The world manages the lifetime of the object.
		*/
		virtual Object* ManageObject(Object* p, std::string name);		
		
		/*! Retrieve an object with a certain name. */
		Object* GetObject(std::string name);
		
		/*! Remove a named Object. */		
		void RemoveObject(std::string name);
										
		/*! Simulate this World by some timestep dt. */
		virtual void Simulate(double dt);

   	/*! Check for collision but deliberately force the particles to
     * collide if they penetrate each other, hence a faster but
     * less accurate algorithm.
     * Detect collision between two generic shapes the call one of
		 * the "Detect_Update()" functions.
     */
		void DetectAndHandleCollision(ColBase* col1, ColBase* col2, double dt, CollisionHandler* ch);
			
		/*! Update the impulse affecting two spheres */
		void ApplyImpulse(CollisionInfo& c1, ColSphere* s1, ColSphere* s2);		
		/*! Update the impulse affecting a plane and a sphere */
		void ApplyImpulse(CollisionInfo& c1, ColPlane* s1, ColSphere* s2);
		/*! Update the impulse affecting a sphere and an oriented face */
		void ApplyImpulse(CollisionInfo& ci, ColSphere* s, ColOrientedFace* o);
		/*! Update the impulse affecting a sphere and a cylinder */
		void ApplyImpulse(CollisionInfo& ci, ColSphere* s, ColCylinder* c);
		/*! Update the impulse affecting a sphere and an AABB */
		void ApplyImpulse(CollisionInfo& ci, ColSphere* s, ColAABB* a);
		

		/*! Separates two spheres that have collided then applies an impulse force */
		void HandleCollision(CollisionInfo& ci, ColSphere* s1, ColSphere* s2);		
		/*! Separates two objects that have collided then applies an impulse force */
		void HandleCollision(CollisionInfo& ci, ColPlane* s1, ColSphere* s2);
		/*! Collision has occurred between a sphere and an oriented face */
		void HandleCollision(CollisionInfo& ci, ColSphere* s, ColOrientedFace* o);
		/*! Collision has occurred between a sphere and a cylinder */
		void HandleCollision(CollisionInfo& ci, ColSphere* s, ColCylinder* c);
		/*! Collision has occurred between a sphere and an AABB */
		void HandleCollision(CollisionInfo& ci, ColSphere* s, ColAABB* a);
		
		/*! Adds an impulse force vector to be applied next step */
		inline void AddImpulse(Particle* p1, Vector3 f);

	};

	/*! A generic class for collision handling routines. */
        class CollisionHandler
        {
		public:
        	
		virtual void HandleCollision(CollisionInfo& ci) = 0;
        };

	class DefaultHandler: public CollisionHandler, public World
	{
		private:
		
		World* parent;

	   	public:

		DefaultHandler(World* w)
		: parent(w) {}

		virtual void HandleCollision(CollisionInfo& ci);
	};
}

#endif
