#ifndef CLOTHY
#define CLOTHY

#include "GL/glut.h"
#include "PSystem.h"
#include "RigidStrut.h"
#include "MPWorld.h"

using namespace Penguin;

struct MyTri
{
	Penguin::Particle* p1;
	Penguin::Particle* p2;
	Penguin::Particle* p3;
};

class MyClothPatch: public Penguin::Object
{
	int m_Width; //in # of vertices

	std::vector<Penguin::DampedSpring*> m_StructuralSprings;
	std::vector<Penguin::DampedSpring*> m_ShearSprings;
	std::vector<Penguin::DampedSpring*> m_BendSprings;	
	std::vector<MyTri> m_Tris; //used to make my life easier...
	
	int m_CurrentDrawMode;
	
	void SetupPatch(double width, double str, double shear, double bend, double mass);
	
	void DrawStructs();
	void DrawShears();
	void DrawPoints();
	void DrawPolygons();
	
	inline Vector3& p(int pos)
	{
		return m_Particles[pos]->GetPos();
	}
	
	public:
	
	/*! Creates a default 10x10 cloth patch. */	
	MyClothPatch();
	
	/*! Creates an mxm cloth patch. With optional total width, structural, shear and bend parameters. */
	MyClothPatch(int m, double width=1.0, double structuralStrength=1000.0, double shearStrength=100.0, double bendStrength=100.0, double mass = 0.8);
	
	~MyClothPatch();
	
	/*! Returns a reference to the particle ptr at the i'th row, j'th column. Returns the Particle at 1,1 if i and j are out of bounds */
	Penguin::Particle*& GetParticle(int i, int j);	
		
	/*! Draws this patch. */
	void Draw();	
	
	void NextDrawMode(); 
	
	/* Define Object interface. */
	virtual void SatisfyConstraints(){};		
	virtual void ApplyForces();
	virtual void SimulateParametricParticles(double t){};
	virtual void Update(double dt){}	
	
	virtual bool HasInternalCollisions(){return false;}
};

class TestPSystem
{
	//Particle System info
	PSystem m_System;	
	MyClothPatch* m_Cloth;
	ColSphere* m_b;

	double w; //width/2 of cube sides
	
	//Width,Height of window
	int m_Width, m_Height;
	
	//Camera/Viewpoint Info
	Vector3 m_Camera;	
	Vector3 m_LookAt;
	
	/* frame data */
	int m_NumFrames;
	double m_FPS;
	int m_MSSinceLastDraw;
	
	/*! The milliseconds passed per redraw.*/
	int m_MSPerFrame;
		
	/*! The update flag. */
	bool m_IsUpdating;
	
	/*! Used to hack GLUT callbacks */
	static TestPSystem* m_Instance;
	
	/* prints a string */
	void PrintString(char* s);
	
	/* For MESA */
	void WindowPos2f( GLfloat x, GLfloat y);
	
	/* Internal Initialising Functions */	
	void InitialiseGlut();
	void InitialiseGL();
	
	public:
	TestPSystem();
	~TestPSystem(){}
	
	void Render();
	void Update();	

	void SetCamera(const Vector3& cam){m_Camera = cam;}
	void SetLookAt(const Vector3& lookat){m_LookAt = lookat;}	
	void Set3dViewport(double aspect);
	void Reshape(int w, int h);
	void Display(void);	
	void RenderGuides();	
	void EnableLighting();	
	void BeginMainLoop(void);
	void EnableTimer(int t);
	void DisplayFPS();
	int Time();
	
	void RotateAroundY(double angle);
	void Zoom(double amount);
		
	void KeyListener(unsigned char key, int x, int y);
	void MouseClicked(int button, int state, int x, int y){};
	void MouseDragged(int x, int y){};
	void MouseMoved(int x, int y){};
			
	static void KEYBOARD_CALLBACK(unsigned char key, int x, int y){m_Instance->KeyListener(key,x,y);}	
	static void DISPLAY_CALLBACK(){m_Instance->Display();}	
	static void RESHAPE_CALLBACK(int w, int h){m_Instance->Reshape(w,h);}
	static void TIMER_CALLBACK(int t)
	{
		if (m_Instance->m_IsUpdating) m_Instance->Update();
		 
		m_Instance->m_NumFrames++;
		
		int dt = m_Instance->Time() - m_Instance->m_MSSinceLastDraw;
		
		if (dt>1000)
		{
			m_Instance->m_FPS = m_Instance->m_NumFrames*1000.0/dt;
			m_Instance->m_NumFrames = 0;
			m_Instance->m_MSSinceLastDraw = m_Instance->Time();
		}
						
		glutPostRedisplay();
		glutTimerFunc(m_Instance->m_MSPerFrame,&TestPSystem::TIMER_CALLBACK,0);
	}
	
	static void MOUSECLICK_CALLBACK(int button, int state, int x, int y)
	{
		m_Instance->MouseClicked(button,state,x,y);
	} 
	
	static void MOUSEMOVE_CALLBACK(int x, int y)
	{
		m_Instance->MouseDragged(x,y);
	}
	
	static void PASSIVEMOUSEMOVE_CALLBACK(int x, int y)
	{
		m_Instance->MouseMoved(x,y);
	}
};


#endif
