#include "fountain.h"
#include "IL/ilut.h"

#include <iostream>

using namespace Penguin;

double speed_factor = 1; //i.e. how much should we update the world relative to the screen

TestPSystem* TestPSystem::m_Instance = NULL;

GLuint rippletex;
GLuint ftop,fside;
GLuint watertex;

float water_u = 0; //used for animating water tex.
float water_v = 0;

MyFountain* mf;

void DrawRipples();
void UpdateRipples();
void DrawFountainBase();

struct Ripple
{
	static const float RATE;
	static const double DT;
	static const double START_SIZE;
	static const double END_SIZE;
	static const double DIFF;
	
	Vector3 _pos;
	float _size;
	float y;

	Ripple(Vector3 pos):_pos(pos),_size(START_SIZE),y(0.01){}
	
  void Draw()
	{
		glBegin(GL_QUADS);
			glTexCoord2f(0,0);
			glVertex3f(_pos[0]-_size,y,_pos[2]-_size);

			glTexCoord2f(1,0);
			glVertex3f(_pos[0]+_size,y,_pos[2]-_size);

			glTexCoord2f(1,1);
			glVertex3f(_pos[0]+_size,y,_pos[2]+_size);

			glTexCoord2f(0,1);
			glVertex3f(_pos[0]-_size,y,_pos[2]+_size);			
		glEnd();		
	}	
	
	bool Update()
	{
		return (_size += RATE*DT) > END_SIZE;
	}
};

class RippleHandler: public CollisionHandler
{
	public:
	
	virtual void HandleCollision(CollisionInfo& ci);
};

const float Ripple::RATE = 0.001; //i.e. size increase per millisecond
const double Ripple::DT = 5;

const double Ripple::START_SIZE = 0.01;
const double Ripple::END_SIZE = 0.2;
const double Ripple::DIFF = END_SIZE-START_SIZE;

#include <deque>
using std::deque;

deque<Ripple> g_ripples;

int main(int argc, char** argv)
{
	glutInit(&argc,argv);
	
	ilInit();
	iluInit();
	ilutRenderer(ILUT_OPENGL);	

	TestPSystem test;
	test.EnableTimer(5);	
	test.BeginMainLoop();	

	
	return 0;
}

TestPSystem::TestPSystem():m_Width(500),m_Height(500), m_NumFrames(0), m_FPS(0), m_MSSinceLastDraw(0), m_MSPerFrame(0), m_IsUpdating(true)
{
	InitialiseGlut();
	InitialiseGL();

	world = m_System.AddWorld("Test", new World(5,0.02));
	if (!world)
	{
		std::cout << "Can't create world!" << std::endl;
	}
	else
	{
		world->ManageObject(mf = new MyFountain(), "Fountain");
		mf->Setup();
		mf->SetBehaviour(new FountainEmitter(Vector3(0,1,0),4,0.2));
		mf->SetLife(3);
		mf->SetRate(0.005);
		mf->SetInternalCollisions(false);

		Object* obj = new Object();
		obj->AddCollisionStruct(new ColPlane(Vector3(0,1,0),Vector3::ZERO));
		world->ManageObject(obj,"Water");

		world->RegisterObjToObjHandler(mf,obj,new RippleHandler());

		SetCamera(Vector3(1.75,2.3,-3.5));
		SetLookAt(Vector3(0,1,0));
		EnableLighting();
	}	
}
	
void TestPSystem::Render(void)
{
		//draw pond
	REAL w = 10;
	glColor3f(1,1,1);

	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, watertex);

	glBegin(GL_QUADS);
		glTexCoord2f(water_u,water_v);
		glVertex3f(-w,0,-w);
		
		glTexCoord2f(water_u+1,water_v);
		glVertex3f(w,0,-w);
		
		glTexCoord2f(water_u+1,water_v+1);
		glVertex3f(w,0,w);
		
		glTexCoord2f(water_u,water_v+1);
		glVertex3f(-w,0,w);		
	glEnd();
	
	glDisable(GL_TEXTURE_2D);

	//Draw fountain reflection
	glEnable(GL_BLEND);
	glBlendEquation(GL_ADD);
	glDisable(GL_DEPTH_TEST);

	glRotatef(180,0,0,1);

	DrawFountainBase();

	glRotatef(180,0,0,1);

	glEnable(GL_DEPTH_TEST);
	glDisable(GL_BLEND);
	
	//draw fountain
	DrawFountainBase();

	//draw fountain spray
	mf->Draw();

	//draw ripples
	DrawRipples();
}

void TestPSystem::Update(void)
{
	static double dt = m_MSPerFrame/1000.0;
	m_System.Simulate(dt*speed_factor);
	UpdateRipples();

	static float U_RATE = 0.05;
	static float V_RATE = 0.02;
	water_u += U_RATE*dt;
	water_v += V_RATE*dt;	
}

void TestPSystem::KeyListener(unsigned char key, int x, int y)
{
	if (key==' ') m_IsUpdating = !m_IsUpdating;	
	if (key=='a') RotateAroundY(0.1);	
	else if (key=='d') RotateAroundY(-0.1);	
	else if (key=='w') Zoom(1.1);
	else if (key=='s') Zoom(0.9);
	else if (key=='r') AddVecToCamera(Vector3(0,0.1,0));
	else if (key=='f') AddVecToCamera(Vector3(0,-0.1,0));
}

/* SET UP GLUT AND OPENGL */
void TestPSystem::InitialiseGlut()
{
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);

	//Create a single window
	glutInitWindowPosition(100,100);
	glutInitWindowSize(m_Width,m_Height);		
	glutCreateWindow("Fountastic!");		
	
	//Callbacks
	glutDisplayFunc(&TestPSystem::DISPLAY_CALLBACK);
	glutReshapeFunc(&TestPSystem::RESHAPE_CALLBACK);
	glutKeyboardFunc(&TestPSystem::KEYBOARD_CALLBACK);	
	glutMouseFunc(&TestPSystem::MOUSECLICK_CALLBACK);
	glutMotionFunc(&TestPSystem::MOUSEMOVE_CALLBACK);
	glutPassiveMotionFunc(&TestPSystem::PASSIVEMOUSEMOVE_CALLBACK);
	
	glutSetCursor(GLUT_CURSOR_CROSSHAIR);
}

void TestPSystem::InitialiseGL()
{
	glClearColor(0,0.2,1,1);	
	double aspect = (double)m_Height/m_Width;	
	Set3dViewport(aspect);
	
	rippletex = ilutGLLoadImage("ripple.png");
	ftop = ilutGLLoadImage("f_top.png");
	fside = ilutGLLoadImage("f_side.png");
	watertex = ilutGLLoadImage("water.jpg");
}

void TestPSystem::Reshape(int w, int h)
{
	m_Width = w;
	m_Height = h;
	
	double aspect = (double)m_Height/m_Width;
		
	glViewport(0,0,w,h);
	
	Set3dViewport(aspect);
}

void TestPSystem::Display(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();
	
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	
	DisplayFPS();	
	
	glEnable(GL_DEPTH_TEST);	
	glEnable(GL_LIGHTING);

	Vector3 lookat;
	lookat = m_LookAt;

	gluLookAt(m_Camera[0],m_Camera[1],m_Camera[2],lookat[0],lookat[1],lookat[2],0,1,0);
	
	Render();
	
	glutSwapBuffers();
}

void TestPSystem::Set3dViewport(double aspect)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60,1/aspect,.1,1000); 
	
	glMatrixMode(GL_MODELVIEW);
}

void TestPSystem::RotateAroundY(double angle)
{
	Vector3 dV;

	dV = m_Camera - m_LookAt;

	double x = dV[0];
	double z = dV[2];
	
	x = dV[0]*cos(angle) + dV[2]*sin(angle);
	z = -dV[0]*sin(angle) + dV[2]*cos(angle);
	
	dV[0] = x;
	dV[2] = z;
	
	m_Camera = dV + m_LookAt;	
}

void TestPSystem::RenderGuides()
{
	glColor3f(0.5,0.5,0.5);
	
	glBegin(GL_LINE_LOOP);
		glVertex3f(-1,-1,-1);
		glVertex3f(1,-1,-1);
		glVertex3f(1,-1,1);
		glVertex3f(-1,-1,1);
	glEnd();
	
	glBegin(GL_LINE_LOOP);
		glVertex3f(-1,1,-1);
		glVertex3f(1,1,-1);
		glVertex3f(1,1,1);
		glVertex3f(-1,1,1);
	glEnd();
	
	glBegin(GL_LINES);
		glVertex3f(-1,-1,-1);
		glVertex3f(-1,1,-1);
		
		glVertex3f(1,-1,-1);
		glVertex3f(1,1,-1);
		
		glVertex3f(1,-1,1);
		glVertex3f(1,1,1);
		
		glVertex3f(-1,-1,1);
		glVertex3f(-1,1,1);		
	glEnd();

}

void TestPSystem::EnableLighting()
{
	GLfloat col[] = {1,1,0.8,1.0};
	GLfloat spec[] = {1,1,1,1};
	GLfloat shine[] = {1.0};
	
	GLfloat pos[] = {-15,10,3,1};
	GLfloat pos2[] = {0,10,5,1};
	
	GLfloat diff[] = {1.0,1.0,1.0,1.0};
	GLfloat lmodel_ambient[] = {0.2,0.2,0.2,1};
	
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, col);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, spec);	
	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shine);	
		
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diff);
	glLightfv(GL_LIGHT0, GL_POSITION, pos);
	
	glLightfv(GL_LIGHT1, GL_DIFFUSE, diff);
	glLightfv(GL_LIGHT1, GL_POSITION, pos2);
	
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
	
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHT1);
	
	glEnable(GL_COLOR_MATERIAL);
}

void TestPSystem::Zoom(double amount)
{
	Vector3 dV = m_Camera - m_LookAt;
	dV = dV/amount;
	m_Camera = m_LookAt + dV;	
}

void TestPSystem::BeginMainLoop(void)
{
	TestPSystem::m_Instance = this;
	glutMainLoop();
}

void TestPSystem::EnableTimer(int t)
{
	m_MSPerFrame = t;
	glutTimerFunc(t,&TestPSystem::TIMER_CALLBACK,0);
}

int TestPSystem::Time()
{
	return glutGet(GLUT_ELAPSED_TIME);
}

void TestPSystem::PrintString(char* s)
{
	static void* font = {GLUT_BITMAP_9_BY_15};

	if (s && std::strlen(s))
	{
		while (*s)
		{
			glutBitmapCharacter((void*)font,*s);
			s++;
		}
	}
}

void TestPSystem::DisplayFPS()
{
	glColor3f(1,1,1);
	WindowPos2f(m_Width-100,m_Height-15);

	char buffer[10];
	std::sprintf(buffer,"fps: %g",m_FPS);	
	PrintString(buffer);
	
	WindowPos2f(m_Width-109,m_Height-30);
	double mspf = 1000*1/m_FPS; //milliseconds per frame
	std::sprintf(buffer,"mspf: %g",mspf);
	PrintString(buffer);
		
	WindowPos2f(m_Width-109,m_Height-45);
	std::sprintf(buffer,"msps: %g",mspf- m_MSPerFrame);
	PrintString(buffer);
}

void TestPSystem::WindowPos2f( GLfloat x, GLfloat y)
{	
	double z = 0;
	double w = 1;

   GLfloat fx, fy;

   /* Push current matrix mode and viewport attributes */
   glPushAttrib( GL_TRANSFORM_BIT | GL_VIEWPORT_BIT );

   /* Setup projection parameters */
   glMatrixMode( GL_PROJECTION );
   glPushMatrix();
   glLoadIdentity();
   glMatrixMode( GL_MODELVIEW );
   glPushMatrix();
   glLoadIdentity();

   glDepthRange( z, z );
   glViewport( (int) x - 1, (int) y - 1, 2, 2 );

   /* set the raster (window) position */
   fx = x - (int) x;
   fy = y - (int) y;
   glRasterPos4f( fx, fy, 0.0, w );

   /* restore matrices, viewport and matrix mode */
   glPopMatrix();
   glMatrixMode( GL_PROJECTION );
   glPopMatrix();

   glPopAttrib();
}

/***************************** FOUNTAIN STUFF ********************************/
/*******************************************************************************/

MyFountain::MyFountain():CollidableEmitter(3000,0.5,0.025)
{
}

void MyFountain::Draw()
{
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE,GL_ONE);
	
	int i;
	for(i=0; i<m_CurrentParticles;++i)
	{
		Vector3& pos = m_Particles[i]->GetPos();
		glTranslatef(pos[0],pos[1],pos[2]);
		glutSolidSphere(0.025,8,8);
		glTranslatef(-pos[0],-pos[1],-pos[2]);
	}	

	glDisable(GL_BLEND);
}

void MyFountain::Kill(Particle* p)
{
	int i;
	for(i=0;i<m_CurrentParticles;++i)
	{
		if (m_Particles[i]==p)
		{
			m_LifeArray[i] = -1;
			return;
		}
	}
}


void DrawRipples()
{
	std::deque<Ripple>::iterator it = g_ripples.begin();
	std::deque<Ripple>::iterator end = g_ripples.end();

	glDisable(GL_LIGHTING);
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE,GL_ONE);

	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D,rippletex);

	for(;it!=end;++it)
	{
		it->Draw();
	}

	glDisable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);
	glEnable(GL_LIGHTING);
}

void UpdateRipples()
{
	std::deque<Ripple>::iterator it = g_ripples.begin();
	std::deque<Ripple>::iterator end = g_ripples.end();

	for(;it!=end;++it)
	{
		if(it->Update()) g_ripples.pop_front();
	}
}

void RippleHandler::HandleCollision(CollisionInfo& ci)
{
	/* We know that target two is our particle. */
	ColSphere* cs = (ColSphere*)ci.target2;
	Vector3 pos = cs->GetTarget()->GetPos();
	Ripple ripple(pos);
	g_ripples.push_back(ripple);

	mf->Kill(cs->GetTarget());	
}

void DrawFountainBase()
{
	//draw fountain
	REAL fw = 0.2;
	REAL h = 1;

	/* The top */
	
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, ftop);

	glNormal3f(0,1,0);
	glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex3f(-fw,h,-fw);
		glTexCoord2f(1,0);
		glVertex3f(fw,h,-fw);
		glTexCoord2f(1,1);
		glVertex3f(fw,h,fw);
		glTexCoord2f(0,1);
		glVertex3f(-fw,h,fw);
	glEnd();
	
	/* The sides */
	
	glBindTexture(GL_TEXTURE_2D, fside);

	int i;
	for(i=0;i<4;++i)
	{
	
		glNormal3f(0,0,1);
		glBegin(GL_QUADS);
			glTexCoord2f(0,0);
			glVertex3f(-fw,h,fw);
			glTexCoord2f(1,0);
			glVertex3f(fw,h,fw);
			glTexCoord2f(1,1);
			glVertex3f(fw,0,fw);
			glTexCoord2f(0,1);
			glVertex3f(-fw,0,fw);
		glEnd();

		glRotatef(90,0,1,0);
	}
	
	glDisable(GL_TEXTURE_2D);


}

