#ifndef DAMPEDSPRING_H
#define DAMPEDSPRING_H

#include "Force.h"

namespace Penguin
{
	/*! \brief DampedSpring represents a damped spring force between two particles.
	 *
	 * A damped spring applies forces to two particles depending on their
	 * relative distance from one another. The damping either represents an
	 * attached dampener or the inter-molecular damping of the spring.
	 *
	 * Note: A large spring constant <b>k</b>, or large damping constant <b>c</b>
	 * will most likely lead to unstable results.
	 */
	 
	class DampedSpring: public Force
	{
		protected:
	
		Particle* m_One;
		Particle* m_Two;

		/*! Rest length. */
		double m_L;
		
		/*! Spring constant. */
		double m_K;
		
		/*! Damping constant. */
		double m_C;	
	
		public:
	
		/*! C'tor */
		DampedSpring(Particle* one, Particle* two, double l = 1.0,double k = 1.0, double c = 0.01);	
		
		/*! Applies a damped spring force between its particles.
		*
		* Applies hookes law on two particles <b>a</b> and <b>b</b>:
		* 
		* \f$ f_a = -k_s (\|L\|-R) + k_d \dot L \cdot \frac{L}{|L|}) \frac{L}{|L|} \\
		*	
		* f_b = -f_a \\
		*	
		* L = a-b \\
		*
		* \dot L = v_a - v_b
		*	\f$
		*/
		virtual void ApplyForce(void);
		
		Particle* GetOne(){return m_One;}
		Particle* GetTwo(){return m_Two;}
	};
}

#endif
