#ifndef OBJECT_H
#define OBJECT_H

#include "World.h"
#include "Particle.h"
#include <vector>
#include "Common.h"
#include "Collision.h"

namespace Penguin
{
	/*! \brief An OBJECT is an abstract collection of Particles, Forces and Constraints.
	*
	* Adding Objects to a World helps delineate the borders between the sets of
	* particles, forces and constraints.
	*
	* Examples of Object could be:
	* - A piece of cloth
	* - A rigid crate
	* - A particle emitter (?)
	*
	* Simply inherit and define this interface to be able to add your object to a world.
	*/
	class Object
	{
		protected:
		
		/*! A list of all particles. */ 
		std::vector<Particle*> m_Particles;
		
		/*! A list of parametrically defined particles. */
		std::vector<Particle*> m_ParametricParticles;
						
		/*! A list of collidable particles - DuCK*/
		std::vector<ColBase*> m_CollisionStructs;

		/*! A list of all forces in this world. */
		std::vector<Force*> m_Forces;		
		
		/*! Constraint list. */
		std::vector<Constraint*> m_Constraints;
				
		/*! Internal collisions? */
		bool m_HasInternalCollisions;
				
		public:
			
		Object():m_HasInternalCollisions(true){}
		virtual ~Object();
				
		virtual void SatisfyConstraints();		
		virtual void ApplyForces();
		virtual void SimulateParametricParticles(double t);		
		
		/*! A miscellaneous update function. Called after the simulation has been stepped. */
		virtual void Update(double dt){}
				
		/*! Add a new Particle to this object
		 *
		 * Note: The object assumes that it will now manage the Particle that
		 * p points to. 
		 */
		virtual Particle* AddParticle(Particle* p);			

		/*! Returns the unmanaged particle vector. */
		std::vector<Particle*>* GetParticles(){return &m_Particles;}
		
		virtual PIter GetParticlesBegin(){return m_Particles.begin();}
		virtual PIter GetParticlesEnd(){return m_Particles.end();}
						
		/*! Add a parametric particle to the simulation.
		 *
		 * Note: The object assumes that it will now manage the Particle that
		 * p points to. 
		 */
		virtual Particle* AddParametricParticle(Particle* p);

		/*! Returns the parametric particle vector. */
		std::vector<Particle*>* GetParametricParticles(){return &m_ParametricParticles;}
		
		/*! Adds a collidable struct to the simulation */
    virtual ColBase* AddCollisionStruct(ColBase* p);

		/* Returns the Collidable Particle Vector */
    std::vector<ColBase*>* GetCollisionStructs(){return &m_CollisionStructs;}
		virtual ColIter GetCollisionStructsBegin(){return m_CollisionStructs.begin();}
		virtual ColIter GetCollisionStructsEnd(){return m_CollisionStructs.end();}

		/*! Add a new force to the system. */
		virtual Force* AddForce(Force* f);
				
		/*! Returns the force vector. */
		std::vector<Force*>* GetForces(){return &m_Forces;}
		
		/*! Add a new constraint to the system. */
		virtual Constraint* AddConstraint(Constraint* f);
		
		/*! Returns the constraint vector. */
		std::vector<Constraint*>* GetConstraints(){return &m_Constraints;}
		
		/*! Returns true if inter object collisions should be handled. */
		virtual bool HasInternalCollisions(){return m_HasInternalCollisions;}
		
		/*! Sets whether or not this object will have internal collisions. */
		virtual void SetInternalCollisions(bool t){m_HasInternalCollisions = t;}
		
	};
}

#endif
