#include "ClothPatch.h"

using namespace Penguin;

ClothPatch::ClothPatch(): m_Width(10), m_CurrentDrawMode(0)
{
	SetupPatch(1.0,1.0,1.0,1.0,0.5);
}

ClothPatch::ClothPatch(int m, double width, double structuralStrength, double shearStrength, double bend,double mass): m_Width(m), m_CurrentDrawMode(0)
{
	SetupPatch(width,structuralStrength,shearStrength, bend,mass);
}

void ClothPatch::SetupPatch(double width, double str, double shear, double bend, double mass)
{
	/* Create a Cloth Patch */
	
		double strc = str/80;
		double shearc = shear/80;
		double bendc = bend/80;		
		
		double dist = width/m_Width;		
				
		int i,j;
		for (i=0; i<m_Width; i++)
		{
			for (j=0; j<m_Width; j++)
			{
				double x = width*((double)j/m_Width-0.5);
				double y = width*((1-(double)i/m_Width)-0.5);
			
				Particle* p = new Particle();			
				p->SetPos(Vector3(x,0,y));
				p->SetMass(mass);

				AddParticle(p);      				
				
				ColSphere* cs = new ColSphere(p,0.005);
				cs->SetCoefficientOfRestitution(0);
				AddCollisionStruct(cs);
			}
		}

		for (i=0; i<m_Width-1; i++)
		{
			for (j=0; j<m_Width-1; j++)
			{
				int row = i*m_Width+j;
			
				Tri one;
				one.p1 = m_Particles[row];
				one.p2 = m_Particles[row+m_Width]; //next row
				one.p3 = m_Particles[row+1];
			
				m_Tris.push_back(one);
				
				Tri two;
				two.p1 = m_Particles[row+m_Width];
				two.p2 = m_Particles[row+m_Width+1];
				two.p3 = m_Particles[row+1];
				
				m_Tris.push_back(two);
			}
		}
	
					
		/* horizontal springs */
		
		for(i=0; i<(m_Width); i++)
		{
			for(j=0; j<(m_Width-1); j++)
			{
				m_StructuralSprings.push_back(new DampedSpring( m_Particles[i*m_Width+j],
				m_Particles[i*m_Width+j+1],dist,str,strc));
	
				if (j<(m_Width-2))
				{
					m_BendSprings.push_back(new DampedSpring( m_Particles[i*m_Width+j],
					m_Particles[i*m_Width+j+2],2*dist,bend,bendc));
				}
			}
		}
		
		/* vertical springs */
		
		for(i=0; i<(m_Width-1); i++)
		{
			for(j=0; j<(m_Width); j++)
			{
				m_StructuralSprings.push_back(new DampedSpring(m_Particles[i*m_Width+j],m_Particles[(i+1)*m_Width+j],dist,str,strc));
				
				if (i<(m_Width-2))
				{
					m_BendSprings.push_back(new DampedSpring( m_Particles[i*m_Width+j],
					m_Particles[(i+2)*m_Width+j],2*dist,bend,bendc));
				}
			}
		}	
		
		/* diagonal springs */
		
		double diag = sqrt(dist*dist);
		
		for(i=0; i<(m_Width-1); i++)
		{
			for(j=0; j<(m_Width-1); j++)
			{
				m_ShearSprings.push_back(new DampedSpring(m_Particles[i*m_Width+j],m_Particles[(i+1)*m_Width+j+1],diag,shear,shearc));
				
				m_ShearSprings.push_back(new DampedSpring(m_Particles[(i+1)*m_Width+j],m_Particles[i*m_Width+j+1],diag,shear,shearc));			
			}
		}
}

ClothPatch::~ClothPatch()
{
	std::vector<Particle*>::iterator pit = m_Particles.begin();
	while(pit!=m_Particles.end())
	{
		delete(*pit);
		pit++;
	}
	
	std::vector<Penguin::DampedSpring*>::iterator it = m_StructuralSprings.begin();
	while(it!=m_StructuralSprings.end())
	{
		delete(*it);
		it++;
	}
	
	it = m_ShearSprings.begin();
	while(it!=m_ShearSprings.end())
	{
		delete(*it);
		it++;
	}
	
	it = m_BendSprings.begin();
	while(it!=m_BendSprings.end())
	{
		delete(*it);
		it++;
	}
}

void ClothPatch::Draw()
{
	switch (m_CurrentDrawMode)
	{
		case 0:
		{
			glPointSize(1);
			glColor3f(1,1,1);
			DrawPoints();
			break;
		}
		
		case 1:
		{
			glPointSize(3);
			glColor3f(1,1,1);
			DrawPoints();
			
			glColor3f(0.3,0.3,0.6);
			DrawStructs();
			break;
		}
		
		case 2:
		{
			glPointSize(3);
			glColor3f(1,1,1);
			DrawPoints();
			
			glColor3f(0.3,0.3,0.6);
			DrawStructs();
			
			glColor3f(0.6,0.0,0.0);
			DrawShears();
			break;			
		}
		
		case 3:
		{
			glColor3f(1,1,0);
			DrawPolygons();
			break;
		}
	}
	
}

Particle*& ClothPatch::GetParticle(int i, int j)
{
	if (i<0||i>=m_Width||j<0||j>=m_Width)
	{
		return m_Particles[0];
	}
	
	return m_Particles[i*m_Width+j];
}

void ClothPatch::ApplyForces()
{
	std::vector<DampedSpring*>::iterator it = m_StructuralSprings.begin();
	
	for(;it!=m_StructuralSprings.end();++it)
	{
		(*it)->ApplyForce();
	}
	
	it = m_ShearSprings.begin();
	
	for(;it!=m_ShearSprings.end();++it)
	{
		(*it)->ApplyForce();
	}
	
	it = m_BendSprings.begin();
	
	for(;it!=m_BendSprings.end();++it)
	{
		(*it)->ApplyForce();
	}
}

void ClothPatch::NextDrawMode()
{
	m_CurrentDrawMode++;
	if (m_CurrentDrawMode>3) m_CurrentDrawMode = 0;
	
}

void ClothPatch::DrawStructs()
{
	glBegin(GL_LINES);
	
	std::vector<Penguin::DampedSpring*>::iterator strit = m_StructuralSprings.begin();
	while(strit!=m_StructuralSprings.end())
	{
		glVertex3dv((*strit)->GetOne()->GetPos().GetData());
		glVertex3dv((*strit)->GetTwo()->GetPos().GetData());
		strit++;
	}
	
	glEnd();
}

void ClothPatch::DrawShears()
{
	glBegin(GL_LINES);
	
	std::vector<Penguin::DampedSpring*>::iterator strit = m_ShearSprings.begin();
	while(strit!=m_ShearSprings.end())
	{
		glVertex3dv((*strit)->GetOne()->GetPos().GetData());
		glVertex3dv((*strit)->GetTwo()->GetPos().GetData());
		strit++;
	}
	
	glEnd();
}

void ClothPatch::DrawPoints()
{
	glBegin(GL_POINTS);

	PIter pit = m_Particles.begin();
	while(pit!=m_Particles.end())
	{
		glVertex3dv((*pit)->GetPos().GetData());
		pit++;
	}	
	
	glEnd();
}

void ClothPatch::DrawPolygons()
{
	std::vector<Tri>::iterator it = m_Tris.begin();
	
	glColor3f(1,1,0);
	
	glBegin(GL_TRIANGLES);
	
	for(;it!=m_Tris.end();++it)
	{
		Vector3 v = (it->p1->GetPos()-it->p2->GetPos()).Cross(it->p2->GetPos()-it->p3->GetPos());
		v.Normalise();	
		glNormal3dv(v.GetData());	
		glVertex3dv(it->p1->GetPos().GetData());
		glVertex3dv(it->p2->GetPos().GetData());
		glVertex3dv(it->p3->GetPos().GetData());
	}
	
	glEnd();
}
