
#include "World.h"
#include "PSystem.h"

#include <iostream>

namespace Penguin
{

	World::World(double grav, double res): m_Gravity(grav), m_Resistance(res),m_CollisionTolerance(0.0005),default_handler(0)
	{
	}

	World::~World()
	{
		ObjIter it;
		for(it=m_Objects.begin();it!=m_Objects.end();++it)
		{
			delete it->second;
		}
		delete default_handler;
	}

	Object* World::ManageObject(Object* p, std::string name)
	{
		m_Objects.insert(std::pair<std::string,Object*>(name,p));
		return p;
	}

	Object* World::GetObject(std::string name)
	{
		ObjIter it = m_Objects.find(name);
		if (it == m_Objects.end()) return 0;
		else return it->second; 
	}

	void World::RemoveObject(std::string name)
	{
		ObjIter it = m_Objects.find(name);
		if (it == m_Objects.end()) return;
		else 
		{
			delete it->second;
			m_Objects.erase(it);
		}
	}

	void World::Simulate(double dt)
	{
		if (!default_handler) default_handler = new DefaultHandler(this);		
		
		ZeroForces();
		ApplyForces();
		ApplyImpulses(dt);
		SatisfyConstraints();
		ApplyEulerStep(dt);
		SimulateParametricParticles(m_Parent->GetTime());
		ApplyCollision(dt);
		UpdateObjects(dt);
	}	
	
	void World::ZeroForces()
	{
		ObjIter itm;
		for(itm=m_Objects.begin();itm!=m_Objects.end();++itm)
		{
			PIter it = itm->second->GetParticlesBegin();
			PIter it2 = itm->second->GetParticlesEnd();
						
			for (;it!=it2;++it)
			{
				ZeroForce(*it);
			}			
		}		
	}
	
	void World::ApplyForces()
	{
		ObjIter it;
		for(it=m_Objects.begin();it!=m_Objects.end();++it)
		{
			it->second->ApplyForces();
		}		
	}
	
	void World::ApplyImpulses(double dt)
	{
		ImpIter it;
		for(it=m_ImpulseVector.begin();it!=m_ImpulseVector.end();++it)
		{
			it->DecreaseForce(dt);
			it->ApplyForce();
		}
		
		m_ImpulseVector.erase(m_ImpulseVector.begin(),m_ImpulseVector.end());
	}	
	
	void World::SatisfyConstraints()
	{
		ObjIter it;
		for(it=m_Objects.begin();it!=m_Objects.end();++it)
		{
			it->second->SatisfyConstraints();
		}
	}
	
	void World::ApplyEulerStep(double dt)
	{
		ObjIter itm;
		for(itm=m_Objects.begin();itm!=m_Objects.end();++itm)
		{
			PIter it = itm->second->GetParticlesBegin();
			PIter it2 = itm->second->GetParticlesEnd();
						
			for (;it!=it2;++it)
			{				
				EulerStep(*it,dt);
			}						
		}
	}
	
	void World::SimulateParametricParticles(double t)
	{
		ObjIter itm;
		for(itm=m_Objects.begin();itm!=m_Objects.end();++itm)
		{
			itm->second->SimulateParametricParticles(t);
		}		
	}	
	
	void World::ZeroForce(Particle* p)
	{
		p->SetForce(Vector3(0,-m_Gravity*p->GetMass(),0));
		p->GetForce() -= p->GetVel()*m_Resistance;
	}
	
	void World::EulerStep(Particle* p,double dt)
	{
		p->GetVel() += p->GetForce()*dt/p->GetMass();		
		p->GetPos() += p->GetVel()*dt;
	}	
	
	void World::UpdateObjects(double dt)
	{
		ObjIter itm;
		for(itm=m_Objects.begin();itm!=m_Objects.end();++itm)
		{
			itm->second->Update(dt);
		}
	}
	
	struct ColInfo
	{
		ColIter itBegin;
		ColIter itEnd;
		Object *parent;
		bool hasInternalCollisions;			
	};
	
	void World::ApplyCollision(double dt)
	{
		/* Fill the Collidable Particles vector */
		
		typedef std::vector<ColInfo> ColInfoVec;
		typedef ColInfoVec::iterator ColInfoIter;
		
		ColInfoVec m_CollidableStructs;
	  
	 	ObjIter objIter = m_Objects.begin();
	 	ObjIter objEnd = m_Objects.end();
	  
		for (;objIter!=objEnd;++objIter)
		{
			ColInfo ci;
			ci.itBegin = objIter->second->GetCollisionStructsBegin();
			ci.itEnd = objIter->second->GetCollisionStructsEnd();
			ci.hasInternalCollisions = objIter->second->HasInternalCollisions();	
			ci.parent = objIter->second;
			m_CollidableStructs.push_back(ci);		
		}

		if (m_CollidableStructs.size()==0) return;
		
		ColInfoIter colIter2, colIter = m_CollidableStructs.begin();
		ColInfoIter colEnd = m_CollidableStructs.end();
		
		/* Iterate over all collidable particles and check/handle collision */
				
		for (;colIter!=colEnd;++colIter)
   		 {
			colIter2 = colIter + 1;
      			for (;colIter2!=colEnd;++colIter2)
      			{
				for (ColIter ci = colIter->itBegin; ci != colIter->itEnd; ++ci)
				{
					for (ColIter ci2 = colIter2->itBegin; ci2 != colIter2->itEnd; ++ci2)
					{
						CollisionHandler* colHandler = default_handler;
						std::map<std::pair<Object*,Object*>,CollisionHandler*>::iterator iter = m_ObjObjCollisionHandlers.find(std::make_pair(colIter->parent,colIter2->parent));
						if (iter!=m_ObjObjCollisionHandlers.end()) 
						{
							colHandler = iter->second;	
						}
								
						DetectAndHandleCollision(*ci,*ci2,dt,colHandler);
					}
				}				
			}
			
			if (colIter->hasInternalCollisions)
			{
				for (ColIter ci = colIter->itBegin; ci != colIter->itEnd; ++ci)
				{
					ColIter ci2 = ci + 1;
					for (; ci2 != colIter->itEnd; ++ci2)
					{
						CollisionHandler* colHandler = default_handler;
						std::map<std::pair<Object*,Object*>,CollisionHandler*>::iterator iter = m_ObjObjCollisionHandlers.find(std::make_pair(colIter->parent,colIter2->parent));
						if (iter!=m_ObjObjCollisionHandlers.end())
						{
							colHandler = iter->second;
						}
								
						DetectAndHandleCollision(*ci,*ci2,dt,colHandler);
					}
				}		
			}
		}		
	}

	void World::DetectAndHandleCollision(ColBase* col1, ColBase* col2, double dt, CollisionHandler* colHandler)
	{
		COL_TYPE type1, type2;

		type1 = col1->GetType();
		type2 = col2->GetType();

		if (type1==COL_SPHERE)
		{
			if (type2==COL_SPHERE)
			{
				//SPHERE VS SPHERE
			
				ColSphere* s1 = static_cast<ColSphere*>(col1);
				ColSphere* s2 = static_cast<ColSphere*>(col2);
				CollisionInfo ci = DetectCollision(s1, s2,dt);
				if (ci.areIntersecting) colHandler->HandleCollision(ci);	
				
			}
			else if (type2==COL_PLANE)
			{
				//SPHERE VS PLANE
			
				ColPlane* s1 = static_cast<ColPlane*>(col2);
				ColSphere* s2 = static_cast<ColSphere*>(col1);
				CollisionInfo ci = DetectCollision(s1, s2);
				if (ci.areIntersecting) colHandler->HandleCollision(ci);	
				
			}
			else if (type2==COL_ORFACE)
			{
				//SPHERE VS ORIENTED FACE

				ColSphere* s1 = static_cast<ColSphere*>(col1);
				ColOrientedFace* s2 = static_cast<ColOrientedFace*>(col2);
				CollisionInfo ci = DetectCollision(s1, s2);
				if (ci.areIntersecting) colHandler->HandleCollision(ci);	
			}
			else if (type2==COL_CYLINDER)
			{
				//SPHERE VS CYLINDER
				ColSphere* s1 = static_cast<ColSphere*>(col1);
				ColOrientedFace* s2 = static_cast<ColOrientedFace*>(col2);
				CollisionInfo ci = DetectCollision(s1, s2);
				if (ci.areIntersecting) colHandler->HandleCollision(ci);	
			}
			else if (type2==COL_AABB)
			{
				//SPHERE VS AABB
				ColSphere* s1 = static_cast<ColSphere*>(col1);
				ColAABB* s2 = static_cast<ColAABB*>(col2);
				CollisionInfo ci = DetectCollision(s1, s2);
				if (ci.areIntersecting) colHandler->HandleCollision(ci);	
			}
		}
		else if (type1==COL_PLANE)
		{
			if (type2==COL_SPHERE)
			{
				//PLANE VS SPHERE
			
				ColPlane* s1 = static_cast<ColPlane*>(col1);
				ColSphere* s2 = static_cast<ColSphere*>(col2);
				CollisionInfo ci = DetectCollision(s1, s2);
				if (ci.areIntersecting) colHandler->HandleCollision(ci);	
			}
		}
		else if (type1==COL_ORFACE)
		{	
			if (type2==COL_SPHERE)
			{
				//ORIENTED FACE VS SPHERE
				
				ColSphere* s1 = static_cast<ColSphere*>(col1);
				ColOrientedFace* s2 = static_cast<ColOrientedFace*>(col2);
				CollisionInfo ci = DetectCollision(s1, s2);
				if (ci.areIntersecting) colHandler->HandleCollision(ci);	
			}
		}
		else if (type1==COL_CYLINDER)
		{	
			if (type2==COL_SPHERE)
			{
				//CYLINDER VS SPHERE
				
				ColSphere* s1 = static_cast<ColSphere*>(col1);
				ColCylinder* s2 = static_cast<ColCylinder*>(col2);
				CollisionInfo ci = DetectCollision(s1, s2);
				if (ci.areIntersecting) colHandler->HandleCollision(ci);	
			}
		}		
		else if (type1==COL_AABB)
		{	
			if (type2==COL_SPHERE)
			{
				//AABB VS SPHERE
				
				ColSphere* s1 = static_cast<ColSphere*>(col2);
				ColAABB* s2 = static_cast<ColAABB*>(col1);
				CollisionInfo ci = DetectCollision(s1, s2);
				if (ci.areIntersecting) colHandler->HandleCollision(ci);	
			}
		}		

		//else perform other collision checks between other types...
	}


	/****HandleCollisions functions****\
	\**********************************/
	//FIXME: This is dumb, there has to be a better way...
	void DefaultHandler::HandleCollision(CollisionInfo& ci)
	{
		ColBase *col1 = ci.target1;
		ColBase *col2 = ci.target2;

		if (ci.type1==COL_SPHERE)
		{
			if (ci.type2==COL_SPHERE)
			{
				//SPHERE VS SPHERE
				ColSphere* s1 = static_cast<ColSphere*>(col1);
				ColSphere* s2 = static_cast<ColSphere*>(col2);
				parent->HandleCollision(ci, s1, s2);	
			}
			else if (ci.type2==COL_PLANE)
			{
				//SPHERE VS PLANE
			
				ColPlane* s1 = static_cast<ColPlane*>(col2);
				ColSphere* s2 = static_cast<ColSphere*>(col1);
				parent->HandleCollision(ci, s1, s2);	
			}
			else if (ci.type2==COL_ORFACE)
			{
				//SPHERE VS ORIENTED FACE

				ColSphere* s1 = static_cast<ColSphere*>(col1);
				ColOrientedFace* s2 = static_cast<ColOrientedFace*>(col2);
				parent->HandleCollision(ci, s1, s2);	
			}
			else if (ci.type2==COL_CYLINDER)
			{
				//SPHERE VS CYLINDER
				ColSphere* s1 = static_cast<ColSphere*>(col1);
				ColOrientedFace* s2 = static_cast<ColOrientedFace*>(col2);
				parent->HandleCollision(ci, s1, s2);	
			}
			else if (ci.type2==COL_AABB)
			{
				//SPHERE VS AABB
				ColSphere* s1 = static_cast<ColSphere*>(col1);
				ColAABB* s2 = static_cast<ColAABB*>(col2);
				parent->HandleCollision(ci, s1, s2);	
			}
		}
		else if (ci.type1==COL_PLANE)
		{
			if (ci.type2==COL_SPHERE)
			{
				//PLANE VS SPHERE
			
				ColPlane* s1 = static_cast<ColPlane*>(col1);
				ColSphere* s2 = static_cast<ColSphere*>(col2);
				parent->HandleCollision(ci, s1, s2);	
			}
		}
		else if (ci.type1==COL_ORFACE)
		{	
			if (ci.type2==COL_SPHERE)
			{
				//ORIENTED FACE VS SPHERE
				
				ColSphere* s1 = static_cast<ColSphere*>(col1);
				ColOrientedFace* s2 = static_cast<ColOrientedFace*>(col2);
				parent->HandleCollision(ci, s1, s2);	
			}
		}
		else if (ci.type1==COL_CYLINDER)
		{	
			if (ci.type2==COL_SPHERE)
			{
				//CYLINDER VS SPHERE
				
				ColSphere* s1 = static_cast<ColSphere*>(col1);
				ColCylinder* s2 = static_cast<ColCylinder*>(col2);
				parent->HandleCollision(ci, s1, s2);	
			}
		}		
		else if (ci.type1==COL_AABB)
		{	
			if (ci.type2==COL_SPHERE)
			{
				//AABB VS SPHERE
				
				ColAABB* s1 = static_cast<ColAABB*>(col1);
				ColSphere* s2 = static_cast<ColSphere*>(col2);
				parent->HandleCollision(ci, s2, s1);	
			}
		}		

		//else perform other collision checks between other types...
	}

	//Handles a Collision between two spheres
	//PRE: ci.areIntersecting = true;
	void World::HandleCollision(CollisionInfo& ci, ColSphere* s1, ColSphere* s2)
	{
		//If the distance between the two particles is negligible
		//and if the two particles are moving towards each other
		if ( (ci.intersectionDistance <= m_CollisionTolerance) && (ci.velocityProjectionOnDist < 0.0) )
		{
			ApplyImpulse(ci, s1, s2);
		}
		else if (ci.intersectionDistance > m_CollisionTolerance)
		{
			//penetration: manually move the particles apart
			Vector3 c = ci.collisionNormal * ci.intersectionDistance;

			if (!s1->IsMoveable())
			{
				s2->GetTarget()->GetPos() -= c;
			}
			else if (!s2->IsMoveable())
			{
				s1->GetTarget()->GetPos() += c;
			}		
			else
			{			
				s1->GetTarget()->GetPos() += c/2;
				s2->GetTarget()->GetPos() -= c/2;
			}
			ApplyImpulse(ci, s1, s2);
		}
	}
	
	//Handles a Collision between two spheres
	//PRE: ci.areIntersecting = true;
	void World::HandleCollision(CollisionInfo& ci, ColPlane* s1, ColSphere* s2)
	{
		//If the distance between the two particles is negligible
		//and if the two particles are moving towards each other
		if ( (ci.intersectionDistance <= m_CollisionTolerance) && (ci.velocityProjectionOnDist < 0.0) )
		{
			ApplyImpulse(ci, s1, s2);
		}
		else if (ci.intersectionDistance > m_CollisionTolerance)
		{
			//penetration: manually move the particles apart
			Vector3 c = ci.collisionNormal * ci.intersectionDistance;
	
			if (s2->IsMoveable())
			{
				s2->GetTarget()->GetPos() += c;
			}		
			
			ApplyImpulse(ci, s1, s2);
		}
	}
	
	void World::HandleCollision(CollisionInfo& ci, ColSphere* s, ColOrientedFace* o)
	{
		//If the distance between the two objects is negligible
		//and if the sphere is moving towards the or. face
		if ( (ci.intersectionDistance <= m_CollisionTolerance) && (ci.velocityProjectionOnDist < 0.0) )
		{
			ApplyImpulse(ci, s, o);
		}
		else if (ci.intersectionDistance > m_CollisionTolerance)
		{
			//penetration: manually move the two objects apart
			Vector3 c = ci.collisionNormal * ci.intersectionDistance;
	
			if (s->IsMoveable())
			{
				s->GetTarget()->GetPos() += c;
			}		
			
			ApplyImpulse(ci, s, o);
		}
	}
	
	//Handle a collision between a sphere and a cylinder
	//PRE: ci.areIntersecting == true;
	//Note: Cylinders are not moveable for the moment
	//Because only the sphere is moveable, this function is also identical 
	//to the function above. FIXME: It will get complex if cylinders are moveable	
	void World::HandleCollision(CollisionInfo& ci, ColSphere* s, ColCylinder* c)
	{
		//If the distance between the two objects is negligible
		//and if the sphere is moving towards the or. face
		if ( (ci.intersectionDistance <= m_CollisionTolerance) && (ci.velocityProjectionOnDist < 0.0) )
		{
			ApplyImpulse(ci, s, c);
		}
		else if (ci.intersectionDistance > m_CollisionTolerance)
		{
			//penetration: manually move the two objects apart
			Vector3 move = ci.collisionNormal * ci.intersectionDistance;
	
			if (s->IsMoveable())
			{
				s->GetTarget()->GetPos() += move;
			}		
			
			ApplyImpulse(ci, s, c);
		}
	}
	
	//Handle a collision between a sphere and an AABB
	//PRE: ci.areIntersecting == true;
	//Because only the sphere is moveable, this function is also identical 
	//to the function above. 	
	void World::HandleCollision(CollisionInfo& ci, ColSphere* s, ColAABB* a)
	{
		//If the distance between the two objects is negligible
		//and if the sphere is moving towards the or. face
		if ( (ci.intersectionDistance <= m_CollisionTolerance) && (ci.velocityProjectionOnDist < 0.0) )
		{
			ApplyImpulse(ci, s, a);
		}
		else if (ci.intersectionDistance > m_CollisionTolerance)
		{
			//penetration: manually move the two objects apart
			Vector3 move = ci.collisionNormal * ci.intersectionDistance;
	
			if (s->IsMoveable())
			{
				s->GetTarget()->GetPos() += move;
			}		
			
			ApplyImpulse(ci, s, a);
		}
	}
	

	/****The ApplyImpulse() functions****\
	\*************************************/

	void World::ApplyImpulse(CollisionInfo& c1, ColSphere* s1, ColSphere* s2)
	{
		Vector3 impulse; //impulse force
		double e = (s1->GetCoefficientOfRestitution() + s2->GetCoefficientOfRestitution())/2;

		Particle* ptcl1 = s1->GetTarget();
		Particle* ptcl2 = s2->GetTarget();

		impulse =	c1.collisionNormal * (-(1+e)*(c1.relativeVelocity.Dot(c1.collisionNormal)))/
		    (1/ptcl1->GetMass()+1/ptcl2->GetMass());
	
		if (!s1->IsMoveable())
		{
			AddImpulse(ptcl2,-impulse*2);		
		}
		else if (!s2->IsMoveable())
		{
			AddImpulse(ptcl1,impulse*2);
		}
		else
		{
			AddImpulse(ptcl1,impulse);
			AddImpulse(ptcl2,-impulse);
		}
	}
	
	void World::ApplyImpulse(CollisionInfo& c1, ColPlane* s1, ColSphere* s2)
	{
		Vector3 impulse; //impulse force
		double e = (s1->GetCoefficientOfRestitution() + s2->GetCoefficientOfRestitution())/2;

		Particle* sphere = s2->GetTarget();
	
		double J = sphere->GetMass()*(-(1+e) * c1.relativeVelocity.Dot(c1.collisionNormal));
		impulse = c1.collisionNormal * J;

		/* Add some friction (just a hack!) */
		Vector3& vel = sphere->GetVel();
		// vel2 is the velocity NOT in the direction of collision
		Vector3 vel2 = vel - c1.collisionNormal*vel.Dot(c1.collisionNormal);
		
		REAL resistance = (s1->GetCOF()+s2->GetCOF())/2;		
		vel2 = vel2*resistance;
		vel -= vel2;

		AddImpulse(sphere,impulse);		
	}
	
	void World::ApplyImpulse(CollisionInfo& ci, ColSphere* s, ColOrientedFace* o)
	{
		Vector3 impulse; //impulse force
		double e = (s->GetCoefficientOfRestitution() + o->GetCoefficientOfRestitution())/2;

		Particle* sphere = s->GetTarget();

		double J = sphere->GetMass()*(-(1+e) * ci.relativeVelocity.Dot(ci.collisionNormal));
		impulse = ci.collisionNormal * J;

		AddImpulse(sphere,impulse);	
	}

	void World::ApplyImpulse(CollisionInfo& ci, ColSphere* s, ColCylinder* c)
	{
		Vector3 impulse; //impulse force
		double e = (s->GetCoefficientOfRestitution() + c->GetCoefficientOfRestitution())/2;

		Particle* sphere = s->GetTarget();

		double J = sphere->GetMass()*(-(1+e) * ci.relativeVelocity.Dot(ci.collisionNormal));
		impulse = ci.collisionNormal * J;

		AddImpulse(sphere,impulse);		
	}
		
	void World::ApplyImpulse(CollisionInfo& ci, ColSphere* s, ColAABB* a)
	{
		Vector3 impulse; //impulse force
		double e = (s->GetCoefficientOfRestitution() + a->GetCoefficientOfRestitution())/2;

		Particle* sphere = s->GetTarget();

		double J = sphere->GetMass()*(-(1+e) * ci.relativeVelocity.Dot(ci.collisionNormal));
		impulse = ci.collisionNormal * J;

		AddImpulse(sphere,impulse);		
	}


	void World::AddImpulse(Particle* p1, Vector3 f)
	{
		ImpulseForce i(p1,f);
		m_ImpulseVector.push_back(i);
	}

	//Add a new element to the CollisionHandlers vectors
	void World::RegisterObjToObjHandler(Object *o1, Object *o2, CollisionHandler *handler)
	{
		//Assume pair<o1,o2> and pair<o2,o1> are the same and so get the same handler		
		std::pair<Object*, Object*> p1(o1, o2);
		std::pair<Object*, Object*> p2(o2, o1);
		//This is to make sure there is only one pair of (o1,o2) and hanlder at any time
		//(regardless of the order of o1, o2.
		//Just erase one...
		m_ObjObjCollisionHandlers.erase(p2);
		//Add the other
		m_ObjObjCollisionHandlers.insert(std::make_pair(p1, handler));
		//It doesn't matter if the two operations above fail
	}

	void World::RegisterObjHandler(Object *o, CollisionHandler * handler)
	{
		m_ObjCollisionHandlers.insert(std::make_pair(o, handler));
	}

}
